from __future__ import annotations

import numpy as np
import pytest
from numpy.typing import ArrayLike

import snaphu


def jaccard_similarity(a: ArrayLike, b: ArrayLike) -> float:
    """
    Compute the Jaccard index of two binary masks.

    Returns the Jaccard similarity coefficient (Intersect-over-Union) of the two input
    boolean arrays. A value of 1 indicates that the two masks are identical. A value of
    0 indicates that the two masks are non-overlapping. Values between 0 and 1 indicate
    partial overlap.

    Parameters
    ----------
    a, b : array_like
        The input binary masks.

    Returns
    -------
    float
        The Jaccard similarity coefficient.
    """
    a = np.asanyarray(a)
    b = np.asanyarray(b)

    if (a.dtype != np.bool_) or (b.dtype != np.bool_):
        errmsg = (
            f"the input arrays must each have boolean dtype, instead got {a.dtype=} and"
            f" {b.dtype=}"
        )
        raise TypeError(errmsg)

    return np.sum(a & b) / np.sum(a | b)  # type: ignore[no-any-return]


def simulate_phase_noise(
    corr: ArrayLike, nlooks: float, *, seed: int | None = None
) -> np.ndarray:
    r"""
    Simulate InSAR phase noise.

    Generate pseudo-random noise samples that approximately match the expected
    distribution of multilooked interferogram phase.

    The resulting samples are zero-mean Gaussian distributed, with variance equal to the
    Cramer-Rao bound of the Maximum Likelihood Estimator for the interferometric phase\
    [1]_. This simple approximation is most accurate for high coherence and large number
    of looks. The true phase distribution is more complicated\ [2]_.

    Parameters
    ----------
    corr : array_like
        The interferometric coherence magnitude.
    nlooks : float
        The number of independent looks used to form the interferogram.
    seed : int or None, optional
        A seed for initializing pseudo-random number generator state. Must be
        nonnegative. If None, then the generator will be initialized randomly. Defaults
        to None.

    Returns
    -------
    numpy.ndarray
        Phase noise samples, in radians.

    References
    ----------
    .. [1] E. Rodriguez, and J. M. Martin, "Theory and design of interferometric
       synthetic aperture radars," IEE Proceedings-F, vol. 139, no. 2, pp. 147-159,
       April 1992.
    .. [2] J. S. Lee, K. W. Hoppel, S. A. Mango, and A. R. Miller, "Intensity and phase
       statistics of multilook polarimetric and interferometric SAR imagery," IEEE
       Trans. Geosci. Remote Sens. 32, 1017-1028 (1994).
    """
    # Setup a pseudo-random number generator.
    rng = np.random.default_rng(seed)

    # Approximate interferometric phase standard deviation using a simple approximation
    # that holds for high coherence / number of looks.
    corr = np.asanyarray(corr)
    sigma = 1.0 / np.sqrt(2.0 * nlooks) * np.sqrt(1.0 - corr**2) / corr

    # Generate zero-mean Gaussian-distributed phase noise samples.
    return rng.normal(scale=sigma)


class TestGrowConnComps:
    def test_single_component(self):
        shape = (128, 129)
        unw = np.zeros(shape=shape, dtype=np.float32)
        corr = np.ones(shape=shape, dtype=np.float32)

        conncomp = snaphu.grow_conncomps(unw, corr, nlooks=1.0)

        np.testing.assert_array_equal(conncomp, 1)
        assert conncomp.shape == unw.shape
        assert conncomp.dtype == np.uint32

    def test_output_parameter(self):
        unw = np.zeros(shape=(128, 128), dtype=np.float32)
        corr = np.ones(unw.shape, dtype=np.float32)

        # Test passing a `conncomp` array as an output parameter.
        conncomp = np.zeros(unw.shape, dtype=np.uint8)
        snaphu.grow_conncomps(unw, corr, nlooks=1.0, conncomp=conncomp)

        np.testing.assert_array_equal(conncomp, 1)

    def test_low_coherence(self):
        # Simulate a diagonal phase ramp.
        y, x = np.ogrid[-3:3:1024j, -3:3:1024j]
        unw = np.pi * (x + y)

        # Simulate a single square-shaped island of high-coherence surrounded by
        # low-coherence pixels.
        corr = np.full(unw.shape, fill_value=0.1, dtype=np.float32)
        corr[256:-256, 256:-256] = 1.0

        # Add simulated phase noise to unwrapped phase.
        nlooks = 100.0
        unw += simulate_phase_noise(corr, nlooks, seed=1234)

        # Grow connected components.
        conncomp = snaphu.grow_conncomps(unw, corr, nlooks=1.0)

        # Create a mask of pixels that are expected to have nonzero connected component
        # label.
        mask = corr > 0.5

        # Check approximate locations of nonzero connected component labels.
        assert jaccard_similarity(conncomp != 0, mask) > 0.9

    # Helper method to check connected component labels for the next few test cases.
    def check_conncomp_4_quadrants(self, conncomp: np.ndarray) -> None:
        # Check unique connected component labels.
        assert set(np.unique(conncomp)) == {0, 1, 2, 3, 4}

        # Check each quadrants' connected component label. Ignore the outermost
        # row/column of each quadrant since the way SNAPHU grows connected components
        # sometimes erodes corners.
        np.testing.assert_array_equal(conncomp[:255, :255], 1)
        np.testing.assert_array_equal(conncomp[:255, -254:], 2)
        np.testing.assert_array_equal(conncomp[-254:, :255], 3)
        np.testing.assert_array_equal(conncomp[-254:, -254:], 4)

    def test_mask(self):
        # Simulate a diagonal phase ramp.
        y, x = np.ogrid[-3:3:512j, -3:3:512j]
        unw = np.pi * (x + y)

        # Sample coherence for an interferogram with no noise.
        corr = np.ones(unw.shape, dtype=np.float32)

        # Create a binary mask that subdivides the array into 4 disjoint quadrants of
        # valid samples separated by a single row & column of invalid samples.
        mask = np.ones(unw.shape, dtype=np.bool_)
        mask[256, :] = False
        mask[:, 256] = False

        # Grow connected components.
        conncomp = snaphu.grow_conncomps(unw, corr, nlooks=1.0, mask=mask)

        self.check_conncomp_4_quadrants(conncomp)

    def test_zero_magnitude(self):
        # Simulate a diagonal phase ramp.
        y, x = np.ogrid[-3:3:512j, -3:3:512j]
        unw = np.pi * (x + y)

        # Sample coherence for an interferogram with no noise.
        corr = np.ones(unw.shape, dtype=np.float32)

        # Create an array of interferogram magnitude data. A single row & column of
        # zero-magnitude pixels subdivides the array into 4 disjoint quadrants of
        # nonzero-magnitude pixels.
        mag = np.ones(unw.shape, dtype=np.float32)
        mag[256, :] = 0.0
        mag[:, 256] = 0.0

        # Grow connected components.
        conncomp = snaphu.grow_conncomps(unw, corr, nlooks=1.0, mag=mag)

        self.check_conncomp_4_quadrants(conncomp)

    def test_nans(self):
        # Simulate a diagonal phase ramp.
        y, x = np.ogrid[-3:3:512j, -3:3:512j]
        unw = np.pi * (x + y)

        # Sample coherence for an interferogram with no noise.
        corr = np.ones(unw.shape, dtype=np.float32)

        # Simulate interferogram magnitudes
        mag = np.ones(unw.shape, dtype=np.float32)

        # Insert a single row & column of NaN values to the magnitude & unwrapped phase
        # data such that each array is subdivided into 4 disjoint quadrants of valid
        # pixels separated by NaNs.
        mag[256, :] = np.nan
        mag[:, 256] = np.nan
        unw[256, :] = np.nan
        unw[:, 256] = np.nan

        # Grow connected components.
        conncomp = snaphu.grow_conncomps(unw, corr, nlooks=1.0, mag=mag)

        self.check_conncomp_4_quadrants(conncomp)

    def test_min_conncomp_frac(self):
        unw = np.zeros(shape=(1000, 1000), dtype=np.float32)
        corr = np.ones(shape=unw.shape, dtype=np.float32)

        # Create a mask containing a single region of valid pixels with area equal to 1%
        # of the total array, surrounded by invalid pixels.
        mask = np.zeros(unw.shape, dtype=np.bool_)
        mask[450:550, 450:550] = True

        # Grow connected components with `min_conncomp_frac` slightly less than 0.01.
        # There should be a single nonzero connected component label.
        eps = 1e-3
        conncomp = snaphu.grow_conncomps(
            unw, corr, nlooks=1.0, mask=mask, min_conncomp_frac=0.01 - eps
        )
        assert set(np.unique(conncomp)) == {0, 1}

        # Grow connected components with `min_conncomp_frac` slightly greater than 0.01.
        # There should be no nonzero connected component labels.
        conncomp = snaphu.grow_conncomps(
            unw, corr, nlooks=1.0, mask=mask, min_conncomp_frac=0.01 + eps
        )
        assert set(np.unique(conncomp)) == {0}

    def test_shape_mismatch(self):
        unw = np.empty(shape=(128, 128), dtype=np.float32)
        corr = np.empty(shape=(128, 129), dtype=np.float32)
        pattern = (
            r"^shape mismatch: corr dataset must have shape \(128, 128\), instead got"
            r" corr.shape=\(128, 129\)"
        )
        with pytest.raises(ValueError, match=pattern):
            snaphu.grow_conncomps(unw, corr, nlooks=100.0)

    def test_bad_unw_dtype(self):
        unw = np.empty(shape=(128, 128), dtype=np.complex64)
        corr = np.empty(unw.shape, dtype=np.float32)
        pattern = r"^unw dataset must be real-valued, instead got dtype=complex64$"
        with pytest.raises(TypeError, match=pattern):
            snaphu.grow_conncomps(unw, corr, nlooks=100.0)

    def test_bad_corr_dtype(self):
        unw = np.empty(shape=(128, 128), dtype=np.float32)
        corr = np.empty(unw.shape, dtype=np.complex64)
        pattern = r"^corr dataset must be real-valued, instead got dtype=complex64$"
        with pytest.raises(TypeError, match=pattern):
            snaphu.grow_conncomps(unw, corr, nlooks=100.0)

    def test_bad_nlooks(self):
        unw = np.empty(shape=(128, 128), dtype=np.float32)
        corr = np.empty(unw.shape, dtype=np.float32)
        pattern = "^nlooks must be >= 1, instead got 0.5$"
        with pytest.raises(ValueError, match=pattern):
            snaphu.grow_conncomps(unw, corr, nlooks=0.5)

    def test_bad_cost(self):
        unw = np.empty(shape=(128, 128), dtype=np.float32)
        corr = np.empty(unw.shape, dtype=np.float32)
        pattern = r"^cost mode must be in \{.*\}, instead got 'asdf'$"
        with pytest.raises(ValueError, match=pattern):
            snaphu.grow_conncomps(unw, corr, nlooks=100.0, cost="asdf")



from __future__ import annotations

import numpy as np

from .io import InputDataset, OutputDataset

__all__ = [
    "check_2d_shapes",
    "check_bool_or_byte_dtype",
    "check_complex_dtype",
    "check_cost_mode",
    "check_dataset_shapes",
    "check_float_dtype",
    "check_integer_dtype",
]


def check_2d_shapes(**shapes: tuple[int, ...]) -> None:
    """
    Ensure that the input tuples are valid 2-D shapes.

    Parameters
    ----------
    **shapes : dict, optional
        Input tuples to check for validity. Inputs must be pairs of positive integers.
        The name of each keyword argument is used to format the error message in case of
        an invalid input.

    Raises
    ------
    ValueError
        If an input had invalid length or contained invalid (negative or zero) values.
    """
    for name, shape in shapes.items():
        if len(shape) != 2:
            errmsg = f"{name} must be a pair of ints, instead got {name}={shape}"
            raise ValueError(errmsg)
        if not all(n >= 1 for n in shape):
            errmsg = (
                f"{name} may not contain negative or zero values, got {name}={shape}"
            )
            raise ValueError(errmsg)


def check_dataset_shapes(
    shape: tuple[int, ...], **datasets: InputDataset | OutputDataset
) -> None:
    """
    Ensure that one or more datasets has a specified shape.

    Parameters
    ----------
    shape : tuple of int
        The expected shape of each dataset.
    **datasets : dict, optional
        Datasets whose shape must be equal to `shape`. The name of each keyword argument
        is used to format the error message in case of a shape mismatch.

    Raises
    ------
    ValueError
        If any dataset had a different shape.
    """
    for name, arr in datasets.items():
        if arr.shape != shape:
            errmsg = (
                f"shape mismatch: {name} dataset must have shape {shape}, instead got"
                f" {name}.shape={arr.shape}"
            )
            raise ValueError(errmsg)


def check_complex_dtype(**datasets: InputDataset | OutputDataset) -> None:
    """
    Ensure that one or more datasets is complex-valued.

    Parameters
    ----------
    **datasets : dict, optional
        Datasets whose datatype must be a complex floating-point type. The name of each
        keyword argument is used to format the error message in case of a different
        datatype.

    Raises
    ------
    TypeError
        If any dataset had a different datatype.
    """
    for name, arr in datasets.items():
        if not np.issubdtype(arr.dtype, np.complexfloating):
            errmsg = (
                f"{name} dataset must be complex-valued, instead got dtype={arr.dtype}"
            )
            raise TypeError(errmsg)


def check_float_dtype(**datasets: InputDataset | OutputDataset) -> None:
    """
    Ensure that one or more datasets is real-valued.

    Parameters
    ----------
    **datasets : dict, optional
        Datasets whose datatype must be a floating-point type. The name of each keyword
        argument is used to format the error message in case of a different datatype.

    Raises
    ------
    TypeError
        If any dataset had a different datatype.
    """
    for name, arr in datasets.items():
        if not np.issubdtype(arr.dtype, np.floating):
            errmsg = (
                f"{name} dataset must be real-valued, instead got dtype={arr.dtype}"
            )
            raise TypeError(errmsg)


def check_integer_dtype(**datasets: InputDataset | OutputDataset) -> None:
    """
    Ensure that one or more datasets is integer-valued.

    Parameters
    ----------
    **datasets : dict, optional
        Datasets whose datatype must be a (signed or unsigned) integer type. The name of
        each keyword argument is used to format the error message in case of a different
        datatype.

    Raises
    ------
    TypeError
        If any dataset had a different datatype.
    """
    for name, arr in datasets.items():
        if not np.issubdtype(arr.dtype, np.integer):
            errmsg = (
                f"{name} dataset must be integer-valued, instead got dtype={arr.dtype}"
            )
            raise TypeError(errmsg)


def check_bool_or_byte_dtype(**datasets: InputDataset | OutputDataset) -> None:
    """
    Ensure that one or more datasets is boolean or 8-bit integer-valued.

    Parameters
    ----------
    **datasets : dict, optional
        Datasets whose datatype must be a boolean or 8-bit (signed or unsigned) integer
        type. The name of each keyword argument is used to format the error message in
        case of a different datatype.

    Raises
    ------
    TypeError
        If any dataset had a different datatype.
    """
    for name, arr in datasets.items():
        dtype = arr.dtype
        if (dtype != np.bool_) and (dtype != np.uint8) and (dtype != np.int8):
            errmsg = (
                f"{name} dataset must be a boolean or 8-bit integer-valued, instead got"
                f" dtype={dtype}"
            )
            raise TypeError(errmsg)


def check_cost_mode(cost: str) -> None:
    """
    Ensure that the input SNAPHU cost mode is valid and supported.

    Parameters
    ----------
    cost : str
        The cost mode.

    Raises
    ------
    NotImplementedError
        If the specified cost mode is not supported.
    ValueError
        If the specified cost mode is not a valid SNAPHU cost option.
    """
    cost_modes = {"defo", "smooth"}

    if cost == "topo":
        errmsg = (
            "'topo' cost mode is not currently supported, cost must be one of"
            f" {cost_modes}"
        )
        raise NotImplementedError(errmsg)

    if cost not in cost_modes:
        errmsg = f"cost mode must be in {cost_modes}, instead got {cost!r}"
        raise ValueError(errmsg)



from __future__ import annotations

import os
import textwrap
from pathlib import Path
from tempfile import mkstemp
from typing import cast, overload

import numpy as np

from ._check import (
    check_2d_shapes,
    check_bool_or_byte_dtype,
    check_complex_dtype,
    check_cost_mode,
    check_dataset_shapes,
    check_float_dtype,
    check_integer_dtype,
)
from ._conncomp import regrow_conncomp_from_unw
from ._snaphu import run_snaphu
from ._util import copy_blockwise, nan_to_zero, scratch_directory
from .io import InputDataset, OutputDataset

__all__ = [
    "unwrap",
]


def normalize_and_validate_tiling_params(
    ntiles: tuple[int, int],
    tile_overlap: int | tuple[int, int],
    nproc: int,
) -> tuple[tuple[int, int], tuple[int, int], int]:
    """
    Normalize and validate inputs related to tiling and multiprocessing.

    Parameters
    ----------
    ntiles : (int, int)
        Number of tiles along the row/column directions.
    tile_overlap : int or (int, int)
        Overlap, in pixels, between neighboring tiles.
    nproc : int
        Maximum number of child processes to spawn for parallel tile unwrapping.

    Returns
    -------
    ntiles : (int, int)
        `ntiles` normalized to a pair of positive integers.
    tile_overlap : (int, int)
        `tile_overlap` normalized to a pair of nonnegative integers.
    nproc : int
        `nproc` as a positive integer.
    """
    # Normalize `ntiles` to a tuple and ensure its contents are two positive-valued
    # integers.
    ntiles = tuple(ntiles)  # type: ignore[assignment]
    check_2d_shapes(ntiles=ntiles)

    # If `tile_overlap` is iterable, ensure it's a tuple. Otherwise, assume it's a
    # single integer.
    try:
        tile_overlap = tuple(tile_overlap)  # type: ignore[arg-type,assignment]
    except TypeError:
        tile_overlap = (tile_overlap, tile_overlap)  # type: ignore[assignment]

    # Convince static type checkers that `tile_overlap` is now a pair of ints.
    tile_overlap = cast(tuple[int, int], tile_overlap)

    # Ensure the contents of `tile_overlap` are two nonnegative integers.
    if len(tile_overlap) != 2:
        errmsg = (
            f"tile_overlap must be an int or pair of ints, instead got {tile_overlap=}"
        )
        raise ValueError(errmsg)
    if not all(n >= 0 for n in tile_overlap):
        errmsg = f"tile_overlap may not contain negative values, got {tile_overlap=}"
        raise ValueError(errmsg)

    # If `nproc` is less than 1, use all available processors. Fall back to serial
    # execution if the number of available processors cannot be determined.
    if nproc < 1:
        nproc = os.cpu_count() or 1

    return ntiles, tile_overlap, nproc


@overload
def unwrap(
    igram: InputDataset,
    corr: InputDataset,
    nlooks: float,
    cost: str = "smooth",
    init: str = "mcf",
    *,
    mask: InputDataset | None = None,
    min_conncomp_frac: float = 0.01,
    phase_grad_window: tuple[int, int] = (7, 7),
    ntiles: tuple[int, int] = (1, 1),
    tile_overlap: int | tuple[int, int] = 0,
    nproc: int = 1,
    tile_cost_thresh: int = 500,
    min_region_size: int = 100,
    single_tile_reoptimize: bool = True,
    regrow_conncomps: bool = True,
    scratchdir: str | os.PathLike[str] | None = None,
    delete_scratch: bool = True,
    unw: OutputDataset,
    conncomp: OutputDataset,
) -> tuple[OutputDataset, OutputDataset]: ...  # pragma: no cover


# If `unw` and `conncomp` aren't specified, return the outputs as two NumPy arrays.
@overload
def unwrap(
    igram: InputDataset,
    corr: InputDataset,
    nlooks: float,
    cost: str = "smooth",
    init: str = "mcf",
    *,
    mask: InputDataset | None = None,
    min_conncomp_frac: float = 0.01,
    phase_grad_window: tuple[int, int] = (7, 7),
    ntiles: tuple[int, int] = (1, 1),
    tile_overlap: int | tuple[int, int] = 0,
    nproc: int = 1,
    tile_cost_thresh: int = 500,
    min_region_size: int = 100,
    single_tile_reoptimize: bool = True,
    regrow_conncomps: bool = True,
    scratchdir: str | os.PathLike[str] | None = None,
    delete_scratch: bool = True,
) -> tuple[np.ndarray, np.ndarray]: ...  # pragma: no cover


def unwrap(  # type: ignore[no-untyped-def]
    igram,
    corr,
    nlooks,
    cost="smooth",
    init="mcf",
    *,
    mask=None,
    min_conncomp_frac=0.01,
    phase_grad_window=(7, 7),
    ntiles=(1, 1),
    tile_overlap=0,
    nproc=1,
    tile_cost_thresh=500,
    min_region_size=100,
    single_tile_reoptimize=True,
    regrow_conncomps=True,
    scratchdir=None,
    delete_scratch=True,
    unw=None,
    conncomp=None,
):
    r"""
    Unwrap an interferogram using SNAPHU.

    Performs 2-D phase unwrapping using the Statistical-Cost, Network-Flow Algorithm for
    Phase Unwrapping (SNAPHU)\ [1]_. The algorithm produces a Maximum a Posteriori (MAP)
    estimate of the unwrapped phase field by (approximately) solving a nonlinear network
    flow optimization problem.

    The outputs include the unwrapped phase and a corresponding array of connected
    component labels. Each connected component is a region of pixels in the solution
    that is believed to have been unwrapped in an internally self-consistent manner\
    [2]_. Each distinct region is assigned a unique positive integer label. Pixels not
    belonging to any component are assigned a label of zero.

    Parameters
    ----------
    igram : snaphu.io.InputDataset
        The input interferogram. A 2-D complex-valued array. Not a Number (NaN) values
        in the array will be replaced with zeros.
    corr : snaphu.io.InputDataset
        The sample coherence magnitude. Must be a floating-point array with the same
        dimensions as the input interferogram. Valid coherence values are in the range
        [0, 1]. NaN values in the array will be replaced with zeros.
    nlooks : float
        The equivalent number of independent looks used to form the sample coherence. An
        estimate of the number of statistically independent samples averaged in the
        multilooked data, taking into account spatial correlation due to
        oversampling/filtering (see `Notes`_).
    cost : {'defo', 'smooth'}, optional
        Statistical cost mode. Defaults to 'smooth'.
    init : {'mst', 'mcf'}, optional
        Algorithm used for initialization of unwrapped phase gradients.
        Supported algorithms include Minimum Spanning Tree ('mst') and Minimum
        Cost Flow ('mcf'). Defaults to 'mcf'.
    mask : snaphu.io.InputDataset or None, optional
        Binary mask of valid pixels. Zeros in this raster indicate interferogram
        pixels that should be masked out. If provided, it must have the same dimensions
        as the input interferogram and boolean or 8-bit integer datatype. Defaults to
        None.
    min_conncomp_frac : float, optional
        Minimum size of a single connected component, as a fraction of the total number
        of pixels in the tile. Defaults to 0.01.
    phase_grad_window : (int, int), optional
        The dimensions, in pixels, of the sliding window used for averaging wrapped
        phase gradients to get the mean non-layover slope, in directions parallel and
        perpendicular to the examined phase difference. This parameter corresponds to
        SNAPHU's `KPARDPSI` and `KPERPDPSI` configuration options. Defaults to (7, 7).
    ntiles : (int, int), optional
        Number of tiles along the row/column directions. If `ntiles` is (1, 1), then the
        interferogram will be unwrapped as a single tile. Increasing the number of tiles
        may improve runtime and reduce peak memory utilization, but may also introduce
        tile boundary artifacts in the unwrapped result. Defaults to (1, 1).
    tile_overlap : int or (int, int), optional
        Overlap, in pixels, between neighboring tiles. Increasing overlap may help to
        avoid phase discontinuities between tiles. If `tile_overlap` is a scalar
        integer, the number of overlapping rows and columns will be the same. Defaults
        to 0.
    nproc : int, optional
        Maximum number of child processes to spawn for parallel tile unwrapping. If
        `nproc` is less than 1, use all available processors. Defaults to 1.
    tile_cost_thresh : int, optional
        Cost threshold to use for determining boundaries of reliable regions
        (dimensionless; scaled according to other cost constants). Larger cost threshold
        implies smaller regions -- safer, but more expensive computationally. Defaults
        to 500.
    min_region_size : int, optional
        Minimum size, in pixels, of a reliable region in tile mode. Defaults to 100.
    single_tile_reoptimize : bool, optional
        If True, after unwrapping with multiple tiles, an additional post-processing
        unwrapping step is performed to re-optimize the unwrapped phase using a single
        tile. This option is disregarded when `ntiles` is (1, 1). It supersedes the
        `regrow_conncomps` option -- if both are enabled, only the single-tile
        re-optimization step will be performed in order to avoid redundant computation.
        Defaults to True.
    regrow_conncomps : bool, optional
        If True, the connected component labels will be re-computed using a single tile
        after first unwrapping with multiple tiles. This option is disregarded when
        `ntiles` is (1, 1). It has no effect if `single_tile_reoptimize` was also
        enabled (since regrowing connected components would be redundant in that case).
        Defaults to True.
    scratchdir : path-like or None, optional
        Scratch directory where intermediate processing artifacts are written.
        If the specified directory does not exist, it will be created. If None,
        a temporary directory will be created as though by
        ``tempfile.TemporaryDirectory()``. Defaults to None.
    delete_scratch : bool, optional
        If True, if a scratch directory was created by this function, it will be
        automatically removed from the file system when the function exits. Otherwise,
        the scratch directory will be preserved. Existing directories are not deleted.
        Defaults to True.
    unw : snaphu.io.OutputDataset or None, optional
        An optional output dataset to store the unwrapped phase, in radians. If
        provided, it must have the same dimensions as the input interferogram and
        floating-point datatype. If None, an output array will be allocated internally.
        Defaults to None.
    conncomp : snaphu.io.OutputDataset or None, optional
        An optional output dataset to store the connected component labels. If provided,
        it must have the same dimensions as the input interferogram and any integer
        datatype. If None, an output array will be allocated internally. Defaults to
        None.

    Returns
    -------
    unw : snaphu.io.OutputDataset
        The output unwrapped phase, in radians.
    conncomp : snaphu.io.OutputDataset
        The output connected component labels.

    Notes
    -----
    An estimate of the equivalent number of independent looks may be obtained by

    .. math:: n_e = k_r k_a \frac{d_r d_a}{\rho_r \rho_a}

    where :math:`k_r` and :math:`k_a` are the number of looks in range and azimuth,
    :math:`d_r` and :math:`d_a` are the (single-look) sample spacing in range and
    azimuth, and :math:`\rho_r` and :math:`\rho_a are the range and azimuth resolution.

    References
    ----------
    .. [1] C. W. Chen and H. A. Zebker, "Two-dimensional phase unwrapping with use of
       statistical models for cost functions in nonlinear optimization," Journal of the
       Optical Society of America A, vol. 18, pp. 338-351 (2001).
    .. [2] C. W. Chen and H. A. Zebker, "Phase unwrapping for large SAR interferograms:
       Statistical segmentation and generalized network models," IEEE Transactions on
       Geoscience and Remote Sensing, vol. 40, pp. 1709-1719 (2002).
    """
    # If the `unw` and/or `conncomp` output datasets were not provided, allocate arrays
    # to store these outputs.
    if unw is None:
        unw = np.zeros(shape=igram.shape, dtype=np.float32)
    if conncomp is None:
        conncomp = np.zeros(shape=igram.shape, dtype=np.uint32)

    if igram.ndim != 2:
        errmsg = f"igram must be 2-D, instead got ndim={igram.ndim}"
        raise ValueError(errmsg)

    # Ensure that input & output datasets have matching shapes.
    check_dataset_shapes(igram.shape, corr=corr, unw=unw, conncomp=conncomp)
    if mask is not None:
        check_dataset_shapes(igram.shape, mask=mask)

    # Ensure that input & output datasets have valid datatypes.
    check_complex_dtype(igram=igram)
    check_float_dtype(unw=unw, corr=corr)
    check_integer_dtype(conncomp=conncomp)
    if mask is not None:
        check_bool_or_byte_dtype(mask=mask)

    if nlooks < 1.0:
        errmsg = f"nlooks must be >= 1, instead got {nlooks}"
        raise ValueError(errmsg)

    check_cost_mode(cost)

    init_methods = {"mst", "mcf"}
    if init not in init_methods:
        errmsg = f"init method must be in {init_methods}, instead got {init!r}"
        raise ValueError(errmsg)

    # `phase_grad_window` must be a pair of positive integers.
    phase_grad_window = tuple(phase_grad_window)
    check_2d_shapes(phase_grad_window=phase_grad_window)

    # Validate inputs related to tiling and coerce them to the expected types.
    ntiles, tile_overlap, nproc = normalize_and_validate_tiling_params(
        ntiles=ntiles, tile_overlap=tile_overlap, nproc=nproc
    )

    with scratch_directory(scratchdir, delete=delete_scratch) as dir_:
        # Create a raw binary file in the scratch directory for the interferogram and
        # copy the input data to it. (`mkstemp` is used to avoid data races in case the
        # same scratch directory was used for multiple SNAPHU processes.)
        _, tmp_igram = mkstemp(dir=dir_, prefix="snaphu.igram.", suffix=".c8")
        tmp_igram_mmap = np.memmap(tmp_igram, dtype=np.complex64, shape=igram.shape)
        copy_blockwise(igram, tmp_igram_mmap, transform=nan_to_zero)
        tmp_igram_mmap.flush()

        # Copy the input coherence data to a raw binary file in the scratch directory.
        _, tmp_corr = mkstemp(dir=dir_, prefix="snaphu.corr.", suffix=".f4")
        tmp_corr_mmap = np.memmap(tmp_corr, dtype=np.float32, shape=corr.shape)
        copy_blockwise(corr, tmp_corr_mmap, transform=nan_to_zero)
        tmp_corr_mmap.flush()

        # If a mask was provided, copy the mask data to a raw binary file in the scratch
        # directory.
        if mask is None:
            tmp_mask = None
        else:
            _, tmp_mask = mkstemp(dir=dir_, prefix="snaphu.mask.", suffix=".u1")
            tmp_mask_mmap = np.memmap(tmp_mask, dtype=np.bool_, shape=mask.shape)
            copy_blockwise(mask, tmp_mask_mmap)
            tmp_mask_mmap.flush()

        # Create files in the scratch directory for SNAPHU outputs.
        _, tmp_unw = mkstemp(dir=dir_, prefix="snaphu.unw.", suffix=".f4")
        _, tmp_conncomp = mkstemp(dir=dir_, prefix="snaphu.conncomp.", suffix=".u4")

        config = textwrap.dedent(
            f"""\
            INFILE {tmp_igram}
            INFILEFORMAT COMPLEX_DATA
            CORRFILE {tmp_corr}
            CORRFILEFORMAT FLOAT_DATA
            OUTFILE {tmp_unw}
            OUTFILEFORMAT FLOAT_DATA
            CONNCOMPFILE {tmp_conncomp}
            CONNCOMPOUTTYPE UINT
            LINELENGTH {igram.shape[1]}
            NCORRLOOKS {nlooks}
            STATCOSTMODE {cost.upper()}
            INITMETHOD {init.upper()}
            MINCONNCOMPFRAC {min_conncomp_frac}
            KPARDPSI {phase_grad_window[0]}
            KPERPDPSI {phase_grad_window[1]}
            NTILEROW {ntiles[0]}
            NTILECOL {ntiles[1]}
            ROWOVRLP {tile_overlap[0]}
            COLOVRLP {tile_overlap[1]}
            NPROC {nproc}
            TILECOSTTHRESH {tile_cost_thresh}
            MINREGIONSIZE {min_region_size}
            """
        )
        if mask is not None:
            config += f"BYTEMASKFILE {tmp_mask}\n"

        # Optionally re-optimize the unwrapped phase using a single tile after
        # unwrapping in tiled mode. This step should have no effect when running in
        # single-tile mode, so always skip it in that case.
        single_tile = ntiles == (1, 1)
        if (not single_tile) and single_tile_reoptimize:
            config += "SINGLETILEREOPTIMIZE TRUE\n"

        # Write config parameters to file.
        _, config_file = mkstemp(dir=dir_, prefix="snaphu.config.", suffix=".txt")
        Path(config_file).write_text(config)

        # Run SNAPHU with the specified parameters.
        run_snaphu(config_file)

        # Optionally regrow connected component labels from the unwrapped phase if
        # SNAPHU was run in tiled mode. This step should have no effect if SNAPHU was
        # previously run in single-tile mode or if single-tile re-optimization was
        # enabled, so always skip it in such cases.
        if (not single_tile) and (not single_tile_reoptimize) and regrow_conncomps:
            # The connected component regrowing step takes as input the unwrapped phase,
            # which is missing magnitude information. The magnitude may be necessary,
            # for example, to detect zero-magnitude pixels which should be masked out
            # (i.e. connected component label set to 0). So compute the interferogram
            # magnitude and pass it as a separate input file.
            _, tmp_mag = mkstemp(dir=dir_, prefix="snaphu.mag.", suffix=".f4")
            tmp_mag_mmap = np.memmap(tmp_mag, dtype=np.float32, shape=igram.shape)
            copy_blockwise(tmp_igram_mmap, tmp_mag_mmap, transform=np.abs)
            tmp_mag_mmap.flush()

            # Re-run SNAPHU to compute new connected components from the unwrapped phase
            # as though in single-tile mode, overwriting the original connected
            # components file.
            regrow_conncomp_from_unw(
                unw_file=tmp_unw,
                corr_file=tmp_corr,
                conncomp_file=tmp_conncomp,
                line_length=igram.shape[1],
                nlooks=nlooks,
                cost=cost,
                mag_file=tmp_mag,
                mask_file=tmp_mask,
                min_conncomp_frac=min_conncomp_frac,
                scratchdir=dir_,
            )

        # Get the output unwrapped phase data.
        tmp_unw_mmap = np.memmap(tmp_unw, dtype=np.float32, shape=unw.shape)
        copy_blockwise(tmp_unw_mmap, unw)

        # Get the output connected component labels.
        tmp_cc_mmap = np.memmap(tmp_conncomp, dtype=np.uint32, shape=conncomp.shape)
        copy_blockwise(tmp_cc_mmap, conncomp)

    return unw, conncomp



from __future__ import annotations

import os
import textwrap
from pathlib import Path
from tempfile import mkstemp
from typing import overload

import numpy as np

from ._check import (
    check_bool_or_byte_dtype,
    check_cost_mode,
    check_dataset_shapes,
    check_float_dtype,
    check_integer_dtype,
)
from ._snaphu import run_snaphu
from ._util import copy_blockwise, nan_to_zero, scratch_directory
from .io import InputDataset, OutputDataset

__all__ = [
    "grow_conncomps",
    "regrow_conncomp_from_unw",
]


def regrow_conncomp_from_unw(
    unw_file: str | os.PathLike[str],
    corr_file: str | os.PathLike[str],
    conncomp_file: str | os.PathLike[str],
    line_length: int,
    nlooks: float,
    cost: str,
    mag_file: str | os.PathLike[str] | None = None,
    mask_file: str | os.PathLike[str] | None = None,
    min_conncomp_frac: float = 0.01,
    scratchdir: str | os.PathLike[str] | None = None,
) -> None:
    """
    Run SNAPHU to regrow connected components from an unwrapped input.

    This is particularly useful if SNAPHU was initially run in tiled mode, resulting in
    a set of connected components that are disjoint across tile boundaries. The
    connected components will be recomputed as though by a single tile, so that
    components are no longer delimited by the extents of each tile.

    This function does not compute a new unwrapped solution.

    Parameters
    ----------
    unw_file : path-like
        The input unwrapped phase file path.
    corr_file : path-like
        The input coherence file path.
    conncomp_file : path-like
        The output connected component labels file path.
    line_length : int
        The line length, in samples, of the input & output arrays.
    nlooks : float
        The equivalent number of independent looks used to form the sample coherence.
    cost : str
        Statistical cost mode.
    mag_file : path-like or None, optional
        The input interferogram magnitude file path. If None, magnitude data is not
        used. Defaults to None.
    mask_file : path-like or None, optional
        An optional file path of a byte mask file. If None, no mask is applied. Defaults
        to None.
    min_conncomp_frac : float, optional
        Minimum size of a single connected component, as a fraction of the total number
        of pixels in the array. Defaults to 0.01.
    scratchdir : path-like or None, optional
        The scratch directory where the config file will be written. If None, a
        default temporary directory is chosen as though by ``tempfile.gettempdir()``.
        Defaults to None.
    """
    # In REGROWCONNCOMPS mode, SNAPHU recomputes the cost arrays (but does not
    # re-unwrap), so we should pass in all parameters necessary to compute costs, but
    # don't need to pass an INITMETHOD, for example.
    config = textwrap.dedent(
        f"""\
        REGROWCONNCOMPS TRUE
        INFILE {os.fspath(unw_file)}
        INFILEFORMAT FLOAT_DATA
        UNWRAPPEDINFILEFORMAT FLOAT_DATA
        CORRFILE {os.fspath(corr_file)}
        CORRFILEFORMAT FLOAT_DATA
        CONNCOMPFILE {os.fspath(conncomp_file)}
        CONNCOMPOUTTYPE UINT
        LINELENGTH {line_length}
        NCORRLOOKS {nlooks}
        STATCOSTMODE {cost.upper()}
        MINCONNCOMPFRAC {min_conncomp_frac}
        """
    )
    if mag_file is not None:
        config += f"MAGFILE {os.fspath(mag_file)}\nMAGFILEFORMAT FLOAT_DATA\n"
    if mask_file is not None:
        config += f"BYTEMASKFILE {os.fspath(mask_file)}\n"

    # Write config parameters to file. The config file should have a descriptive name to
    # disambiguate it from the config file used for unwrapping.
    _, config_file = mkstemp(
        dir=scratchdir, prefix="snaphu-regrow-conncomps.config.", suffix=".txt"
    )
    Path(config_file).write_text(config)

    # Run SNAPHU in REGROWCONNCOMPS mode to generate new connected component labels.
    run_snaphu(config_file)


@overload
def grow_conncomps(
    unw: InputDataset,
    corr: InputDataset,
    nlooks: float,
    cost: str = "smooth",
    *,
    mag: InputDataset | None = None,
    mask: InputDataset | None = None,
    min_conncomp_frac: float = 0.01,
    scratchdir: str | os.PathLike[str] | None = None,
    delete_scratch: bool = True,
    conncomp: OutputDataset,
) -> OutputDataset: ...  # pragma: no cover


# If `conncomp` is not specified, return the output as a NumPy array.
@overload
def grow_conncomps(
    unw: InputDataset,
    corr: InputDataset,
    nlooks: float,
    cost: str = "smooth",
    *,
    mag: InputDataset | None = None,
    mask: InputDataset | None = None,
    min_conncomp_frac: float = 0.01,
    scratchdir: str | os.PathLike[str] | None = None,
    delete_scratch: bool = True,
) -> np.ndarray: ...  # pragma: no cover


def grow_conncomps(  # type: ignore[no-untyped-def]
    unw,
    corr,
    nlooks,
    cost="smooth",
    *,
    mag=None,
    mask=None,
    min_conncomp_frac=0.01,
    scratchdir=None,
    delete_scratch=True,
    conncomp=None,
):
    r"""
    Compute connected component labels using SNAPHU.

    Grow connected component labels for the input unwrapped phase field using a
    statistical segmentation framework\ [1]_. Connected components are generated as
    though the phase was unwrapped by the Statistical-Cost, Network-Flow Algorithm for
    Phase Unwrapping (SNAPHU) as a single tile with the specified inputs. The unwrapped
    phase itself is not modified or re-computed.

    This may be useful if the unwrapped phase was formed using a different unwrapping
    algorithm or if the phase was modified after unwrapping by some post-processing
    step(s).

    Each connected component is a region of phase samples that is believed to have been
    unwrapped in an internally self-consistent manner. Each distinct region is assigned
    a unique positive integer label. Pixels not belonging to any component are assigned
    a label of zero.

    Parameters
    ----------
    unw : snaphu.io.InputDataset
        The input unwrapped phase, in radians. A 2-D real-valued array. If the dataset
        contains Not a Number (NaN) values, they will be replaced with zeros.
    corr : snaphu.io.InputDataset
        The sample coherence magnitude. Must be a floating-point array with the same
        dimensions as the input unwrapped phase. Valid coherence values are in the range
        [0, 1]. NaN values in the array will be replaced with zeros.
    nlooks : float
        The equivalent number of independent looks used to form the sample coherence. An
        estimate of the number of statistically independent samples averaged in the
        multilooked data, taking into account spatial correlation due to
        oversampling/filtering (see `Notes`_).
    cost : {'defo', 'smooth'}, optional
        Statistical cost mode. Defaults to 'smooth'.
    mag : snaphu.io.InputDataset or None, optional
        An optional array of interferogram magnitude data. If provided, it must be a
        real-valued array with the same dimensions as the input unwrapped phase.
        Zero-magnitude pixels are treated as invalid (i.e. masked-out). NaN values in
        the array will be replaced with zeros. Defaults to None.
    mask : snaphu.io.InputDataset or None, optional
        Binary mask of valid pixels. Zeros in this raster indicate interferogram
        pixels that should be masked out. If provided, it must have the same dimensions
        as the input unwrapped phase and boolean or 8-bit integer datatype. Defaults to
        None.
    min_conncomp_frac : float, optional
        Minimum size of a single connected component, as a fraction of the total number
        of pixels in the array. Defaults to 0.01.
    scratchdir : path-like or None, optional
        Scratch directory where intermediate processing artifacts are written.
        If the specified directory does not exist, it will be created. If None,
        a temporary directory will be created as though by
        ``tempfile.TemporaryDirectory()``. Defaults to None.
    delete_scratch : bool, optional
        If True, if a scratch directory was created by this function, it will be
        automatically removed from the file system when the function exits. Otherwise,
        the scratch directory will be preserved. Existing directories are not deleted.
        Defaults to True.
    conncomp : snaphu.io.OutputDataset or None, optional
        An optional output dataset to store the connected component labels. If provided,
        it must have the same dimensions as the input unwrapped phase and any integer
        datatype. If None, an output array will be allocated internally. Defaults to
        None.

    Returns
    -------
    conncomp : snaphu.io.OutputDataset
        The output connected component labels.

    Notes
    -----
    An estimate of the equivalent number of independent looks may be obtained by

    .. math:: n_e = k_r k_a \frac{d_r d_a}{\rho_r \rho_a}

    where :math:`k_r` and :math:`k_a` are the number of looks in range and azimuth,
    :math:`d_r` and :math:`d_a` are the (single-look) sample spacing in range and
    azimuth, and :math:`\rho_r` and :math:`\rho_a are the range and azimuth resolution.

    References
    ----------
    .. [1] C. W. Chen and H. A. Zebker, "Phase unwrapping for large SAR interferograms:
       Statistical segmentation and generalized network models," IEEE Transactions on
       Geoscience and Remote Sensing, vol. 40, pp. 1709-1719 (2002).
    """
    # If a `conncomp` output dataset was not provided, create an array to store the
    # results.
    if conncomp is None:
        conncomp = np.zeros(shape=unw.shape, dtype=np.uint32)

    if unw.ndim != 2:
        errmsg = f"unw must be 2-D, instead got ndim={unw.ndim}"
        raise ValueError(errmsg)

    # Ensure that input & output datasets have matching shapes.
    check_dataset_shapes(unw.shape, corr=corr, conncomp=conncomp)
    if mag is not None:
        check_dataset_shapes(unw.shape, mag=mag)
    if mask is not None:
        check_dataset_shapes(unw.shape, mask=mask)

    # Ensure that input & output datasets have valid datatypes.
    check_float_dtype(unw=unw, corr=corr)
    check_integer_dtype(conncomp=conncomp)
    if mag is not None:
        check_float_dtype(mag=mag)
    if mask is not None:
        check_bool_or_byte_dtype(mask=mask)

    if nlooks < 1.0:
        errmsg = f"nlooks must be >= 1, instead got {nlooks}"
        raise ValueError(errmsg)

    check_cost_mode(cost)

    with scratch_directory(scratchdir, delete=delete_scratch) as dir_:
        # Create a raw binary file in the scratch directory for the unwrapped phase and
        # copy the input data to it. (`mkstemp` is used to avoid data races in case the
        # same scratch directory was used for multiple SNAPHU processes.)
        _, tmp_unw = mkstemp(dir=dir_, prefix="snaphu.unw.", suffix=".f4")
        tmp_unw_mmap = np.memmap(tmp_unw, dtype=np.float32, shape=unw.shape)
        copy_blockwise(unw, tmp_unw_mmap, transform=nan_to_zero)
        tmp_unw_mmap.flush()

        # Copy the input coherence data to a raw binary file in the scratch directory.
        _, tmp_corr = mkstemp(dir=dir_, prefix="snaphu.corr.", suffix=".f4")
        tmp_corr_mmap = np.memmap(tmp_corr, dtype=np.float32, shape=corr.shape)
        copy_blockwise(corr, tmp_corr_mmap, transform=nan_to_zero)
        tmp_corr_mmap.flush()

        # If magnitude data was provided, copy it to a raw binary file in the scratch
        # directory.
        if mag is None:
            tmp_mag = None
        else:
            _, tmp_mag = mkstemp(dir=dir_, prefix="snaphu.mag.", suffix=".f4")
            tmp_mag_mmap = np.memmap(tmp_mag, dtype=np.float32, shape=mag.shape)
            copy_blockwise(mag, tmp_mag_mmap, transform=nan_to_zero)
            tmp_mag_mmap.flush()

        # If a mask was provided, copy the mask data to a raw binary file in the scratch
        # directory.
        if mask is None:
            tmp_mask = None
        else:
            _, tmp_mask = mkstemp(dir=dir_, prefix="snaphu.mask.", suffix=".u1")
            tmp_mask_mmap = np.memmap(tmp_mask, dtype=np.bool_, shape=mask.shape)
            copy_blockwise(mask, tmp_mask_mmap)
            tmp_mask_mmap.flush()

        # Create a raw file in the scratch directory for the output connected
        # components.
        _, tmp_conncomp = mkstemp(dir=dir_, prefix="snaphu.conncomp.", suffix=".u4")

        regrow_conncomp_from_unw(
            unw_file=tmp_unw,
            corr_file=tmp_corr,
            conncomp_file=tmp_conncomp,
            line_length=unw.shape[1],
            nlooks=nlooks,
            cost=cost,
            mag_file=tmp_mag,
            mask_file=tmp_mask,
            min_conncomp_frac=min_conncomp_frac,
            scratchdir=dir_,
        )

        # Get the output connected component labels.
        tmp_cc_mmap = np.memmap(tmp_conncomp, dtype=np.uint32, shape=conncomp.shape)
        copy_blockwise(tmp_cc_mmap, conncomp)

    return conncomp
